from texas_hold_em_utils.hands import HandOfTwo
from texas_hold_em_utils.preflop_stats_repository import PreflopStatsRepository
from texas_hold_em_utils.relative_ranking import get_hand_rank_details


class Player:
    hand_of_two = None
    hand_of_five = None
    chips = 0
    round_bet = 0
    in_round = True
    position = -1

    def __init__(self, position, chips=1000):
        """
        Initializes a player with a position (0 to n-1) and chips
        :param position: int from 0 to n-1
        :param chips: int representing the number of chips the player starts with (default 1000)
        """
        self.position = position
        self.hand_of_two = HandOfTwo([])
        self.chips = chips
        self.round_bet = 0
        self.in_round = True

    def bet(self, amount):
        """
        Bets the given amount if the player has enough chips, otherwise bets all chips
        :param amount: amount the player wants to bet
        :return: the amount the player actually bets
        """
        if amount > self.chips:
            amount = self.chips
        self.chips -= amount
        self.round_bet += amount
        return amount

    def fold(self):
        """
        Folds the player's hand and marks them as out of the round
        :return: the player's round bet (0)
        """
        self.in_round = False
        return 0

    def decide(self, round_num, pot, all_day, big_blind, community_cards, player_ct):
        """
        Abstract method for deciding what to do in a round
        :param round_num: 0 for pre-flop, 1 for flop, 2 for turn, 3 for river
        :param pot: the current pot
        :param all_day: the current highest bet (including all rounds)
        :param big_blind: the big blind for the game
        :param community_cards: the community cards (list of 0 to 5 cards)
        :param player_ct: number of players in the game
        :return: a tuple of the action ("fold", "check", "call", "raise") and the amount to bet
        """
        pass


# Simple player calls big blind, then checks, folds to any bet past BB
class SimplePlayer(Player):
    def decide(self, round_num, pot, all_day, big_blind, community_cards, player_ct):
        """
        Simple player calls big blind, then checks, folds to any bet past BB
        :param round_num: 0 for pre-flop, 1 for flop, 2 for turn, 3 for river
        :param pot: the current pot
        :param all_day: the current highest bet (including all rounds)
        :param big_blind: the big blind for the game
        :param community_cards: the community cards (list of 0 to 5 cards)
        :param player_ct: number of players in the game
        :return: a tuple of the action ("fold", "check", "call", "raise") and the amount to bet
        """
        to_call = all_day - self.round_bet
        if round_num == 0 and all_day == big_blind and to_call > 0:
            return "call", self.bet(to_call)
        if to_call == 0:
            return "check", 0
        return "fold", self.fold()


class AllInPreFlopPlayer(Player):
    threshold = 0.5
    preflop_stats_repo = None
    in_check_fold = False

    def __init__(self, position, chips=1000, threshold=0.5):
        super().__init__(position, chips)
        self.threshold = threshold
        self.preflop_stats_repo = PreflopStatsRepository()

    def decide(self, round_num, pot, all_day, big_blind, community_cards, player_ct):
        """
        Goes all in preflop if win rate is above threshold, otherwise check/folds
        :param round_num: 0 for pre-flop, 1 for flop, 2 for turn, 3 for river
        :param pot: the current pot
        :param all_day: the current highest bet (including all rounds)
        :param big_blind: the big blind for the game
        :param community_cards: the community cards (list of 0 to 5 cards)
        :param player_ct: number of players in the game
        :return: a tuple of the action ("fold", "check", "call", "raise") and the amount to bet
        """
        to_call = all_day - self.round_bet
        # if player is BB and
        if self.in_check_fold and to_call > 0 and round_num > 0:
            return "fold", self.fold()
        if round_num == 0:
            self.in_check_fold = False
            win_rate = self.preflop_stats_repo.get_win_rate(
                self.hand_of_two.cards[0].rank,
                self.hand_of_two.cards[1].rank,
                self.hand_of_two.cards[0].suit == self.hand_of_two.cards[1].suit,
                player_ct)
            if win_rate['win_rate'] > self.threshold:
                return "raise", self.bet(self.chips)
            else:
                if to_call == 0:
                    self.in_check_fold = True
                    return "check", 0
                return "fold", self.fold()
        # if player is BB and made it past the flop we need to check/fold
        if self.in_check_fold and to_call > 0 and round_num > 0:
            return "fold", self.fold()
        return "check", 0


class LimpPlayer(Player):
    threshold = 0.5
    preflop_stats_repo = None
    is_variable_threshold = False

    def __init__(self, position, chips=1000, threshold=0.5):
        super().__init__(position, chips)
        self.threshold = threshold
        self.preflop_stats_repo = PreflopStatsRepository()
        if type(threshold) is list:
            self.is_variable_threshold = True

    def decide(self, round_num, pot, all_day, big_blind, community_cards, player_ct):
        """
        Calls if win rate is above threshold, otherwise check/folds
        :param round_num: 0 for pre-flop, 1 for flop, 2 for turn, 3 for river
        :param pot: the current pot
        :param all_day: the current highest bet (including all rounds)
        :param big_blind: the big blind for the game
        :param community_cards: the community cards (list of 0 to 5 cards)
        :param player_ct: number of players in the game
        :return: a tuple of the action ("fold", "check", "call", "raise") and the amount to bet
        """
        to_call = all_day - self.round_bet
        if to_call == 0 or self.chips == 0:
            return "check", 0

        rd_threshold = self.threshold if not self.is_variable_threshold else self.threshold[round_num]

        if round_num == 0:
            win_rate = self.preflop_stats_repo.get_win_rate(
                self.hand_of_two.cards[0].rank,
                self.hand_of_two.cards[1].rank,
                self.hand_of_two.cards[0].suit == self.hand_of_two.cards[1].suit,
                player_ct)
            if win_rate['win_rate'] > rd_threshold:
                return "call", self.bet(to_call)
        else:
            stats = get_hand_rank_details(self.hand_of_two.cards, community_cards, player_ct)
            rate = stats['expected_win_rate']
            if rate > rd_threshold:
                return "call", self.bet(to_call)

        return "fold", self.fold()



import os

import pandas as pd
from dotenv import load_dotenv, find_dotenv
import sqlalchemy


class PostflopStatsRepository:
    all_data = None

    def __init__(self):
        """
        Constructor for the PreflopStatsRepository
        otherwise gets the data from the csv file. If you don't know what to set this to, leave it as False.
        """
        self.all_data = pd.read_csv(os.path.dirname(__file__) + "/data/post_flop_win_rate_distribution.csv")

    def get_percentile(self, win_rate, player_count, street):
        """
        Gets the percentile for the given win rate, player count, and street
        """
        count = self.all_data.query(f"player_ct == {player_count} and street == '{street}' and win_rate < {win_rate}").shape[0]
        total = self.all_data.query(f"player_ct == {player_count} and street == '{street}'").shape[0]
        return count / total



from texas_hold_em_utils.card import Card

import random


class Deck:
    def __init__(self):
        """
        Initializes a deck of 52 cards (does not shuffle)
        """
        self.cards = []
        for suit in [0, 1, 2, 3]:
            for value in range(0, 13):
                self.cards.append(Card().from_ints(value, suit))

    def shuffle(self):
        """
        Shuffles the deck randomly
        :return:
        """
        random.shuffle(self.cards)

    def draw(self):
        """
        Removes a card from the deck and returns it
        :return: the card drawn
        """
        return self.cards.pop()

    def remove(self, card):
        """
        Removes a specific card from the deck
        :param card: the card to be removed
        :return:
        """
        self.cards.remove(card)

    def __len__(self):
        return len(self.cards)

    def __str__(self):
        return str([str(card) for card in self.cards])



import pytest

from texas_hold_em_utils.card import Card
from texas_hold_em_utils.preflop_stats_repository import PreflopStatsRepository
from texas_hold_em_utils.relative_ranking import get_hand_rank_details, rank_hand_post_flop, rank_hand_post_turn, \
    rank_hand_post_river


def test_get_hand_rank_details_missing_hand_card():
    hand = [Card().from_ints(0, 0)]
    # should throw exception
    with pytest.raises(ValueError):
        get_hand_rank_details(hand)

def test_get_hand_rank_details_too_many_hand_cards():
    hand = [Card().from_ints(0, 0), Card().from_ints(0, 1), Card().from_ints(0, 2)]
    # should throw exception
    with pytest.raises(ValueError):
        get_hand_rank_details(hand)

def test_get_hand_rank_details_too_many_community_cards():
    hand = [Card().from_ints(0, 0), Card().from_ints(0, 1)]
    community_cards = [Card().from_ints(0, 2),
                       Card().from_ints(0, 3),
                       Card().from_ints(3, 3),
                       Card().from_ints(3, 2),
                       Card().from_ints(3, 1),
                       Card().from_ints(7, 1)]
    # should throw exception
    with pytest.raises(ValueError):
        get_hand_rank_details(hand, community_cards)

def test_get_hand_rank_details_preflop_2_player():
    hand = [Card().from_ints(0, 0), Card().from_ints(0, 1)]
    community_cards = []

    rank_details = get_hand_rank_details(hand, community_cards)

    assert rank_details["expected_win_rate"] == PreflopStatsRepository().get_win_rate(0, 0, False, 2)['win_rate']
    assert rank_details["expected_win_rate"] == rank_details["expected_2_player_win_rate"]
    assert rank_details["percentile"] > 0.0
    assert rank_details["percentile"] < 100.0


def test_get_hand_rank_details_preflop_3_player():
    hand = [Card().from_ints(0, 0), Card().from_ints(0, 1)]
    community_cards = []

    rank_details = get_hand_rank_details(hand, community_cards, player_count=3)

    assert rank_details["expected_win_rate"] == PreflopStatsRepository().get_win_rate(0, 0, False, 3)['win_rate']
    assert rank_details["expected_win_rate"] < rank_details["expected_2_player_win_rate"]
    assert rank_details["percentile"] > 0.0
    assert rank_details["percentile"] < 100.0


def test_get_hand_rank_details_postflop_2_player():
    hand = [Card().from_ints(0, 0), Card().from_ints(0, 1)]
    community_cards = [
        Card().from_ints(3, 2),
        Card().from_ints(3, 1),
        Card().from_ints(7, 1)
    ]

    rank_details = get_hand_rank_details(hand, community_cards, player_count=2)

    # allow a 15% tollerance to avoid random failure due to sample variability
    assert rank_details["expected_win_rate"] * 1.15 > rank_details["expected_2_player_win_rate"]
    assert rank_details["expected_win_rate"] * 0.85 < rank_details["expected_2_player_win_rate"]
    assert rank_details["percentile"] > 0.0
    assert rank_details["percentile"] < 100.0


def test_get_hand_rank_details_postflop_3_player():
    hand = [Card().from_ints(0, 0), Card().from_ints(0, 1)]
    community_cards = [
        Card().from_ints(3, 2),
        Card().from_ints(3, 1),
        Card().from_ints(7, 1)
    ]

    rank_details = get_hand_rank_details(hand, community_cards, player_count=3)

    # allow a 20% tollerance to avoid random failure due to sample variability
    assert rank_details["expected_win_rate"] * 1.2 > rank_hand_post_flop(hand, community_cards, n_other_players=2)
    assert rank_details["expected_win_rate"] * 0.8 < rank_hand_post_flop(hand, community_cards, n_other_players=2)
    assert rank_details["expected_win_rate"] < rank_details["expected_2_player_win_rate"]
    assert rank_details["percentile"] > 0.0
    assert rank_details["percentile"] < 100.0


def test_get_hand_rank_details_postturn_2_player():
    hand = [Card().from_ints(0, 0), Card().from_ints(0, 1)]
    community_cards = [
        Card().from_ints(3, 2),
        Card().from_ints(3, 1),
        Card().from_ints(7, 1),
        Card().from_ints(8, 1)
    ]

    rank_details = get_hand_rank_details(hand, community_cards, player_count=2)

    # allow a 15% tollerance to avoid random failure due to sample variability
    assert rank_details["expected_win_rate"] * 1.15 > rank_details["expected_2_player_win_rate"]
    assert rank_details["expected_win_rate"] * 0.85 < rank_details["expected_2_player_win_rate"]
    assert rank_details["percentile"] > 0.0
    assert rank_details["percentile"] < 100.0


def test_get_hand_rank_details_postturn_3_player():
    hand = [Card().from_ints(0, 0), Card().from_ints(0, 1)]
    community_cards = [
        Card().from_ints(3, 2),
        Card().from_ints(3, 1),
        Card().from_ints(7, 1),
        Card().from_ints(8, 1)
    ]

    rank_details = get_hand_rank_details(hand, community_cards, player_count=3)

    # allow a 15% tollerance to avoid random failure due to sample variability
    assert rank_details["expected_win_rate"] * 1.15 > rank_hand_post_turn(hand, community_cards, n_other_players=2)
    assert rank_details["expected_win_rate"] * 0.85 < rank_hand_post_turn(hand, community_cards, n_other_players=2)
    assert rank_details["expected_win_rate"] < rank_details["expected_2_player_win_rate"]
    assert rank_details["percentile"] > 0.0
    assert rank_details["percentile"] < 100.0


def test_get_hand_rank_details_postriver_2_player():
    hand = [Card().from_ints(0, 0), Card().from_ints(0, 1)]
    community_cards = [
        Card().from_ints(3, 2),
        Card().from_ints(3, 1),
        Card().from_ints(7, 1),
        Card().from_ints(8, 1),
        Card().from_ints(9, 1)
    ]

    rank_details = get_hand_rank_details(hand, community_cards, player_count=2)

    # allow a 15% tollerance to avoid random failure due to sample variability
    assert rank_details["expected_win_rate"] == rank_details["expected_2_player_win_rate"]
    assert rank_details["percentile"] > 0.0
    assert rank_details["percentile"] < 100.0


def test_get_hand_rank_details_postriver_3_player():
    hand = [Card().from_ints(0, 0), Card().from_ints(0, 1)]
    community_cards = [
        Card().from_ints(3, 2),
        Card().from_ints(3, 1),
        Card().from_ints(7, 1),
        Card().from_ints(8, 1),
        Card().from_ints(9, 1)
    ]

    rank_details = get_hand_rank_details(hand, community_cards, player_count=3)

    # allow a 15% tollerance to avoid random failure due to sample variability
    assert rank_details["expected_win_rate"] == (rank_hand_post_river(hand, community_cards) ** 2)
    assert rank_details["expected_win_rate"] < rank_details["expected_2_player_win_rate"]
    assert rank_details["percentile"] > 0.0
    assert rank_details["percentile"] < 100.0



import copy
from texas_hold_em_utils.deck import Deck
from texas_hold_em_utils.hands import HandOfFive
from texas_hold_em_utils.post_flop_stats_repository import PostflopStatsRepository
from texas_hold_em_utils.preflop_stats_repository import PreflopStatsRepository
from scipy.stats import norm

preflop_stats_repository = PreflopStatsRepository()
postflop_stats_repository = PostflopStatsRepository()

# Source: https://github.com/amarkules1/texas-holdem-notebooks/blob/main/percentiles_from_win_rates.ipynb
percentile_standard_deviations = {
    "flop": 0.20277154473723782,
    "turn": 0.23895325562434003,
    "river": 0.29411125409761246
}


def get_hand_rank_details(hand, community_cards=None, player_count=2):
    """
    Calculates win rate and percentile for a given hand at any point in the game
    :param hand: array of 2 cards
    :param community_cards: array of 3-5 cards or None for pre-flop (default None)
    :param player_count: number of players in the game (default 2)
    :return: a dict of:
        {
        "expected_win_rate": float between 0 and 1, based on the player_count,
        "expected_2_player_win_rate": float between 0 and 1, assumes 2 players,
        "percentile": float between 0 and 100, how the hand's win rate compares to possible hands for other players
        }
    """
    expected_win_rate = 0.0
    expected_2_player_win_rate = 0.0
    percentile = 0.0

    # validate params
    if len(hand) != 2:
        raise ValueError("hand must contain exactly 2 cards")
    if community_cards is not None and len(community_cards) > 5:
        raise ValueError("there can only be up to 5 community cards in texas holdem")
    if player_count < 2:
        raise ValueError("player_count must be at least 2")

    # preflop
    if community_cards is None or len(community_cards) < 3:
        n_player_data = preflop_stats_repository.get_win_rate(hand[0].rank, hand[1].rank, hand[0].suit == hand[1].suit,
                                                              player_count)
        if player_count > 2:
            two_player_data = preflop_stats_repository.get_win_rate(hand[0].rank, hand[1].rank,
                                                                    hand[0].suit == hand[1].suit, 2)
        else:
            two_player_data = n_player_data
        expected_win_rate = n_player_data["win_rate"]
        expected_2_player_win_rate = two_player_data["win_rate"]
        percentile = n_player_data["percentile"]

    # flop 
    if community_cards is not None and len(community_cards) == 3:
        expected_2_player_win_rate = rank_hand_post_flop(hand, community_cards)

        if player_count > 2:
            expected_win_rate = rank_hand_post_flop(hand, community_cards, n_other_players=(player_count - 1))
            percentile = postflop_stats_repository.get_percentile(expected_win_rate, player_count, 'flop')
        else:
            expected_win_rate = expected_2_player_win_rate
            percentile = expected_percentile(expected_2_player_win_rate, 0.5, percentile_standard_deviations['flop'])
    # turn
    if community_cards is not None and len(community_cards) == 4:
        expected_2_player_win_rate = rank_hand_post_turn(hand, community_cards)
        if player_count > 2:
            expected_win_rate = rank_hand_post_turn(hand, community_cards, n_other_players=(player_count - 1))
            percentile = postflop_stats_repository.get_percentile(expected_win_rate, player_count, 'turn')
        else:
            expected_win_rate = expected_2_player_win_rate
            percentile = expected_percentile(expected_2_player_win_rate, 0.5, percentile_standard_deviations['turn'])
    # river
    if community_cards is not None and len(community_cards) == 5:
        expected_2_player_win_rate = rank_hand_post_river(hand, community_cards)
        expected_win_rate = expected_2_player_win_rate ** (player_count - 1)
        if player_count > 2:
            percentile = postflop_stats_repository.get_percentile(expected_win_rate, player_count, 'river')
        else:
            percentile = expected_percentile(expected_2_player_win_rate, 0.5, percentile_standard_deviations['river'])

    return {
        "expected_win_rate": expected_win_rate,
        "expected_2_player_win_rate": expected_2_player_win_rate,
        "percentile": percentile
    }


def rank_hand_post_river(hand, community_cards):
    """
    Calculates the expected win rate for a given hand post-river in a Texas Hold'em game.

    Assumes 2 players. To get the win rate for more than 2 players,
    raise the returned value to the number of players minus 1.
    Seems to work based on: https://github.com/amarkules1/texas-holdem-notebooks/blob/main/n_player_win_rates.ipynb

    :param hand: a list of 2 Card objects representing the player's hand
    :param community_cards: a list of 4 Card objects representing the flop, turn, and river
    :return: float: the expected win rate for the given hand
    """
    if len(hand) != 2 or len(community_cards) != 5:
        raise ValueError("Invalid input: hand and community_cards must be lists of 2 and 5 Card objects respectively.")

    player_hand = HandOfFive(hand, community_cards)
    deck1 = Deck()
    deck2 = Deck()
    wins = 0
    losses = 0
    ties = 0
    for card1 in deck1.cards:
        if card1 not in hand + community_cards:
            for card2 in deck2.cards:
                if card2 not in hand + community_cards and card2 != card1:
                    other_hand = HandOfFive([card1, card2], community_cards)
                    if player_hand > other_hand:
                        wins += 1
                    elif player_hand < other_hand:
                        losses += 1
                    else:
                        ties += 1
    return (wins + (0.5 * ties)) / (wins + losses + ties)


def rank_hand_post_turn(hand, flop_and_turn, n_other_players=1, sample_size=1000):
    """
    Calculates the expected win rate for a given hand post-turn in a Texas Hold'em game.

    :param hand: a list of 2 Card objects representing the player's hand
    :param flop_and_turn: a list of 4 Card objects representing the flop and turn
    :param n_other_players: number of other players, excluding the player whose hand is being analyzed (default 1)
    :param sample_size: number of simulation runs (default 1000)
    :return: float: the expected win rate for the given hand
    """
    # Validate params
    if len(hand) != 2 or len(flop_and_turn) != 4:
        raise ValueError("Invalid input: hand and flop_and_turn must be lists of 2 and 4 Card objects respectively.")

    deck = Deck()
    deck.remove(hand[0])
    deck.remove(hand[1])

    deck.remove(flop_and_turn[0])
    deck.remove(flop_and_turn[1])
    deck.remove(flop_and_turn[2])
    deck.remove(flop_and_turn[3])

    wins = 0.0

    for i in range(sample_size):
        deck_cpy = copy.deepcopy(deck)
        deck_cpy.shuffle()

        river = deck_cpy.draw()

        comm = [flop_and_turn[0], flop_and_turn[1], flop_and_turn[2], flop_and_turn[3], river]

        other_hands = []
        for j in range(n_other_players):
            other_hands.append(HandOfFive([deck_cpy.draw(), deck_cpy.draw()], comm))

        player_hand = HandOfFive(hand, comm)

        lost_to = []
        tied_with = []

        for hand_obj in other_hands:
            if player_hand < hand_obj:
                lost_to.append(hand_obj)
            elif player_hand == hand_obj:
                tied_with.append(hand_obj)

        if len(lost_to) == 0:
            wins += 1 / (len(tied_with) + 1)

    return wins / sample_size


def rank_hand_post_flop(hand, flop, n_other_players=1, sample_size=1000):
    """
    Calculates the expected win rate for a given hand post-flop in a Texas Hold'em game.

    :param hand: a list of 2 Card objects representing the player's hand
    :param flop: a list of 3 Card objects representing the flop and turn
    :param n_other_players: number of other players, excluding the player whose hand is being analyzed (default 1)
    :param sample_size: number of simulation runs (default 1000)
    :return: float: the expected win rate for the given hand
    """
    # Validate params
    if len(hand) != 2 or len(flop) != 3:
        raise ValueError("Invalid input: hand and flop must be lists of 2 and 3 Card objects respectively.")

    deck = Deck()
    deck.remove(hand[0])
    deck.remove(hand[1])

    deck.remove(flop[0])
    deck.remove(flop[1])
    deck.remove(flop[2])

    wins = 0.0

    for i in range(sample_size):
        deck_cpy = copy.deepcopy(deck)
        deck_cpy.shuffle()

        turn = deck_cpy.draw()
        river = deck_cpy.draw()

        comm = [flop[0], flop[1], flop[2], turn, river]

        other_hands = []
        for j in range(n_other_players):
            other_hands.append(HandOfFive([deck_cpy.draw(), deck_cpy.draw()], comm))

        player_hand = HandOfFive(hand, comm)

        lost_to = []
        tied_with = []

        for hand_obj in other_hands:
            if player_hand < hand_obj:
                lost_to.append(hand_obj)
            elif player_hand == hand_obj:
                tied_with.append(hand_obj)

        if len(lost_to) == 0:
            wins += 1 / (len(tied_with) + 1)

    return wins / sample_size


def expected_percentile(win_rate, mean, std_dev):
    """
    Calculates the expected percentile for a given win rate based on a normal distribution.

    :param win_rate: the expected win rate (e.g., 0.25 for a 25% chance of winning)
    :param mean: the mean of the normal distribution (e.g., 0.5 )
    :param std_dev: the standard deviation of the normal distribution (e.g., 0.2 for a 20% std. deviation)
    :return: the expected percentile (e.g., 10.0 for a win rate in the 10th percentile)
    """
    z_score = (win_rate - mean) / std_dev
    return norm.cdf(z_score) * 100




import pytest
from texas_hold_em_utils.post_flop_stats_repository import PostflopStatsRepository

@pytest.fixture
def repo():
    return PostflopStatsRepository()
def test_get_percentile_flop_3_players(repo):
    percentile = repo.get_percentile(0.4806666666666667, 3, 'flop')
    assert isinstance(percentile, float)
    assert 0 <= percentile <= 1

def test_get_percentile_turn_4_players(repo):
    percentile = repo.get_percentile(0.5, 4, 'turn')
    assert isinstance(percentile, float)
    assert 0 <= percentile <= 1

def test_get_percentile_edge_cases(repo):
    # Test with minimum win rate
    min_percentile = repo.get_percentile(0.0, 3, 'flop')
    assert isinstance(min_percentile, float)
    assert min_percentile == 0.0

    # Test with maximum win rate
    max_percentile = repo.get_percentile(1.0, 3, 'flop')
    assert isinstance(max_percentile, float)
    assert max_percentile > 0.999

def test_get_percentile_different_player_counts(repo):
    for player_count in [3, 4, 5, 6]:
        percentile = repo.get_percentile(0.5, player_count, 'flop')
        assert isinstance(percentile, float)
        assert 0 <= percentile <= 1

def test_get_percentile_all_streets(repo):
    for street in ['flop', 'turn', 'river']:
        percentile = repo.get_percentile(0.5, 3, street)
        assert isinstance(percentile, float)
        assert 0 <= percentile <= 1



from texas_hold_em_utils.card import Card
from texas_hold_em_utils.game_utils import get_card_counts, find_royal_flush, find_straight_flush, find_four_of_a_kind, \
    find_full_house, find_flush, find_straight, find_three_of_a_kind, find_two_pair, find_single_pair, find_high_card


def test_get_card_counts_all_diff():
    hand = [Card().from_str("2", "Hearts"), Card().from_str("A", "Spades")]
    community_cards = [Card().from_str("J", "Clubs"), Card().from_str("5", "Diamonds"), Card().from_str("10", "Hearts"),
                       Card().from_str("7", "Spades"), Card().from_str("8", "Clubs")]

    assert get_card_counts(hand, community_cards) == [1, 0, 0, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1]


def test_get_card_counts_2_pair():
    hand = [Card().from_str("2", "Hearts"), Card().from_str("J", "Spades")]
    community_cards = [Card().from_str("J", "Clubs"), Card().from_str("5", "Diamonds"), Card().from_str("10", "Hearts"),
                       Card().from_str("7", "Spades"), Card().from_str("2", "Clubs")]

    assert get_card_counts(hand, community_cards) == [2, 0, 0, 1, 0, 1, 0, 0, 1, 2, 0, 0, 0]


def test_get_card_counts_3_of_a_kind():
    hand = [Card().from_str("2", "Hearts"), Card().from_str("J", "Spades")]
    community_cards = [Card().from_str("J", "Clubs"), Card().from_str("5", "Diamonds"), Card().from_str("10", "Hearts"),
                       Card().from_str("J", "Spades"), Card().from_str("2", "Clubs")]

    assert get_card_counts(hand, community_cards) == [2, 0, 0, 1, 0, 0, 0, 0, 1, 3, 0, 0, 0]


def test_find_royal_flush_true():
    hand = [Card().from_str("10", "Hearts"),
            Card().from_str("J", "Hearts")]
    community_cards = [Card().from_str("Q", "Hearts"),
                       Card().from_str("K", "Hearts"),
                       Card().from_str("A", "Hearts"),
                       Card().from_str("9", "Hearts"),
                       Card().from_str("3", "Spades")]

    assert find_royal_flush(hand, community_cards) == [Card().from_str("A", "Hearts"),
                                                       Card().from_str("K", "Hearts"),
                                                       Card().from_str("Q", "Hearts"),
                                                       Card().from_str("J", "Hearts"),
                                                       Card().from_str("10", "Hearts")]


def test_find_royal_flush_mismatched_suite():
    hand = [Card().from_str("10", "Hearts"),
            Card().from_str("J", "Hearts")]
    community_cards = [Card().from_str("Q", "Hearts"),
                       Card().from_str("K", "Spades"),
                       Card().from_str("A", "Hearts"),
                       Card().from_str("9", "Hearts"),
                       Card().from_str("3", "Spades")]

    assert not find_royal_flush(hand, community_cards)


def test_find_royal_flush_missing_facecard():
    hand = [Card().from_str("4", "Hearts"),
            Card().from_str("J", "Hearts")]
    community_cards = [Card().from_str("Q", "Hearts"),
                       Card().from_str("K", "Hearts"),
                       Card().from_str("A", "Hearts"),
                       Card().from_str("9", "Hearts"),
                       Card().from_str("3", "Spades")]

    assert not find_royal_flush(hand, community_cards)


def test_find_straight_flush():
    hand = [Card().from_str("4", "Hearts"),
            Card().from_str("5", "Hearts")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("K", "Spades"),
                       Card().from_str("A", "Hearts"),
                       Card().from_str("2", "Hearts"),
                       Card().from_str("3", "Hearts")]

    assert find_straight_flush(hand, community_cards) == [Card().from_str("6", "Hearts"),
                                                          Card().from_str("5", "Hearts"),
                                                          Card().from_str("4", "Hearts"),
                                                          Card().from_str("3", "Hearts"),
                                                          Card().from_str("2", "Hearts")]


def test_find_straight_flush_missing_card():
    hand = [Card().from_str("4", "Hearts"),
            Card().from_str("9", "Hearts")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("K", "Spades"),
                       Card().from_str("A", "Hearts"),
                       Card().from_str("2", "Hearts"),
                       Card().from_str("3", "Hearts")]

    assert not find_straight_flush(hand, community_cards)


def test_find_straight_flush_wrong_suite():
    hand = [Card().from_str("4", "Diamonds"),
            Card().from_str("5", "Hearts")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("K", "Spades"),
                       Card().from_str("A", "Hearts"),
                       Card().from_str("2", "Hearts"),
                       Card().from_str("3", "Hearts")]

    assert not find_straight_flush(hand, community_cards)


def test_find_four_of_a_kind():
    hand = [Card().from_str("4", "Hearts"),
            Card().from_str("4", "Spades")]
    community_cards = [Card().from_str("4", "Clubs"),
                       Card().from_str("4", "Diamonds"),
                       Card().from_str("3", "Hearts"),
                       Card().from_str("2", "Hearts"),
                       Card().from_str("3", "Spades")]

    assert find_four_of_a_kind(hand, community_cards) == [Card().from_str("4", "Hearts"),
                                                          Card().from_str("4", "Diamonds"),
                                                          Card().from_str("4", "Clubs"),
                                                          Card().from_str("4", "Spades"),
                                                          Card().from_str("3", "Hearts")]


def test_find_four_of_a_kind_not_found():
    hand = [Card().from_str("4", "Hearts"),
            Card().from_str("6", "Spades")]
    community_cards = [Card().from_str("4", "Clubs"),
                       Card().from_str("4", "Diamonds"),
                       Card().from_str("3", "Hearts"),
                       Card().from_str("2", "Hearts"),
                       Card().from_str("3", "Spades")]

    assert not find_four_of_a_kind(hand, community_cards)


def test_find_full_house():
    hand = [Card().from_str("4", "Hearts"),
            Card().from_str("4", "Spades")]
    community_cards = [Card().from_str("4", "Clubs"),
                       Card().from_str("3", "Diamonds"),
                       Card().from_str("3", "Hearts"),
                       Card().from_str("2", "Hearts"),
                       Card().from_str("6", "Spades")]

    assert find_full_house(hand, community_cards) == [Card().from_str("4", "Hearts"),
                                                      Card().from_str("4", "Spades"),
                                                      Card().from_str("4", "Clubs"),
                                                      Card().from_str("3", "Diamonds"),
                                                      Card().from_str("3", "Hearts")]


def test_find_full_house_no_pair():
    hand = [Card().from_str("4", "Hearts"),
            Card().from_str("4", "Spades")]
    community_cards = [Card().from_str("4", "Clubs"),
                       Card().from_str("7", "Diamonds"),
                       Card().from_str("8", "Hearts"),
                       Card().from_str("2", "Hearts"),
                       Card().from_str("6", "Spades")]

    assert not find_full_house(hand, community_cards)


def test_find_flush():
    hand = [Card().from_str("4", "Hearts"),
            Card().from_str("9", "Hearts")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("K", "Spades"),
                       Card().from_str("A", "Hearts"),
                       Card().from_str("2", "Hearts"),
                       Card().from_str("3", "Hearts")]

    assert find_flush(hand, community_cards) == [Card().from_str("A", "Hearts"),
                                                 Card().from_str("9", "Hearts"),
                                                 Card().from_str("6", "Hearts"),
                                                 Card().from_str("4", "Hearts"),
                                                 Card().from_str("3", "Hearts")]


def test_find_flush_no_flush():
    hand = [Card().from_str("4", "Hearts"),
            Card().from_str("9", "Spades")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("K", "Spades"),
                       Card().from_str("A", "Hearts"),
                       Card().from_str("2", "Hearts"),
                       Card().from_str("3", "Spades")]

    assert not find_flush(hand, community_cards)


def test_find_full_house_no_three_of_a_kind():
    hand = [Card().from_str("4", "Hearts"),
            Card().from_str("4", "Spades")]
    community_cards = [Card().from_str("5", "Clubs"),
                       Card().from_str("2", "Diamonds"),
                       Card().from_str("3", "Hearts"),
                       Card().from_str("2", "Hearts"),
                       Card().from_str("6", "Spades")]

    assert not find_full_house(hand, community_cards)


def test_find_straight():
    hand = [Card().from_str("2", "Hearts"),
            Card().from_str("5", "Spades")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("7", "Spades"),
                       Card().from_str("8", "Hearts"),
                       Card().from_str("9", "Spades"),
                       Card().from_str("J", "Hearts")]

    assert find_straight(hand, community_cards) == [Card().from_str("9", "Spades"),
                                                    Card().from_str("8", "Hearts"),
                                                    Card().from_str("7", "Spades"),
                                                    Card().from_str("6", "Hearts"),
                                                    Card().from_str("5", "Spades")]


def test_find_straight_no_straight():
    hand = [Card().from_str("2", "Hearts"),
            Card().from_str("5", "Spades")]
    community_cards = [Card().from_str("Q", "Hearts"),
                       Card().from_str("7", "Spades"),
                       Card().from_str("8", "Hearts"),
                       Card().from_str("9", "Spades"),
                       Card().from_str("J", "Hearts")]

    assert not find_straight(hand, community_cards)


def test_find_three_of_a_kind():
    hand = [Card().from_str("5", "Hearts"),
            Card().from_str("5", "Spades")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("7", "Spades"),
                       Card().from_str("5", "Clubs"),
                       Card().from_str("9", "Spades"),
                       Card().from_str("J", "Hearts")]

    assert find_three_of_a_kind(hand, community_cards) == [Card().from_str("5", "Hearts"),
                                                           Card().from_str("5", "Spades"),
                                                           Card().from_str("5", "Clubs"),
                                                           Card().from_str("J", "Hearts"),
                                                           Card().from_str("9", "Spades")]


def test_find_three_of_a_kind_no_three_of_a_kind():
    hand = [Card().from_str("5", "Hearts"),
            Card().from_str("5", "Spades")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("7", "Spades"),
                       Card().from_str("8", "Clubs"),
                       Card().from_str("9", "Spades"),
                       Card().from_str("J", "Hearts")]

    assert not find_three_of_a_kind(hand, community_cards)


def test_find_two_pair():
    hand = [Card().from_str("A", "Hearts"),
            Card().from_str("2", "Spades")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("A", "Spades"),
                       Card().from_str("2", "Clubs"),
                       Card().from_str("9", "Spades"),
                       Card().from_str("J", "Hearts")]

    assert find_two_pair(hand, community_cards) == [Card().from_str("A", "Hearts"),
                                                    Card().from_str("A", "Spades"),
                                                    Card().from_str("2", "Spades"),
                                                    Card().from_str("2", "Clubs"),
                                                    Card().from_str("J", "Hearts")]


def test_find_two_pair_no_two_pair():
    hand = [Card().from_str("A", "Hearts"),
            Card().from_str("2", "Spades")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("A", "Spades"),
                       Card().from_str("3", "Clubs"),
                       Card().from_str("9", "Spades"),
                       Card().from_str("J", "Hearts")]

    assert not find_two_pair(hand, community_cards)


def test_find_single_pair():
    hand = [Card().from_str("A", "Hearts"),
            Card().from_str("2", "Spades")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("A", "Spades"),
                       Card().from_str("3", "Clubs"),
                       Card().from_str("9", "Spades"),
                       Card().from_str("J", "Hearts")]

    assert find_single_pair(hand, community_cards) == [Card().from_str("A", "Hearts"),
                                                       Card().from_str("A", "Spades"),
                                                       Card().from_str("J", "Hearts"),
                                                       Card().from_str("9", "Spades"),
                                                       Card().from_str("6", "Hearts")]


def test_find_single_pair_no_pair():
    hand = [Card().from_str("A", "Hearts"),
            Card().from_str("2", "Spades")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("3", "Spades"),
                       Card().from_str("4", "Clubs"),
                       Card().from_str("9", "Spades"),
                       Card().from_str("J", "Hearts")]

    assert not find_single_pair(hand, community_cards)


def test_find_high_card_hand():
    hand = [Card().from_str("A", "Hearts"),
            Card().from_str("2", "Spades")]
    community_cards = [Card().from_str("6", "Hearts"),
                       Card().from_str("3", "Spades"),
                       Card().from_str("4", "Clubs"),
                       Card().from_str("9", "Spades"),
                       Card().from_str("J", "Hearts")]

    assert find_high_card(hand, community_cards) == [Card().from_str("A", "Hearts"),
                                                     Card().from_str("J", "Hearts"),
                                                     Card().from_str("9", "Spades"),
                                                     Card().from_str("6", "Hearts"),
                                                     Card().from_str("4", "Clubs")]



import os

import pandas as pd
from dotenv import load_dotenv, find_dotenv
import sqlalchemy


class PreflopStatsRepository:
    conn = None
    use_database = False
    all_data = None

    def __init__(self, use_database=False):
        """
        Constructor for the PreflopStatsRepository
        :param use_database: Default False. gets the data from the database if True and DATABASE_CONN_STRING env variable is set,
        otherwise gets the data from the csv file. If you don't know what to set this to, leave it as False.
        """
        self.use_database = use_database
        if use_database and os.getenv("DATABASE_CONN_STRING"):
            _ = load_dotenv(find_dotenv())
            self.conn = sqlalchemy.create_engine(os.getenv("DATABASE_CONN_STRING")).connect()
        else:
            self.all_data = pd.read_csv(os.path.dirname(__file__) + "/data/preflop_win_rates.csv")

    def get_win_rate(self, card1_rank, card2_rank, suited, player_count):
        """
        Gets win rate and related info for the given cards and player count
        :param card1_rank: 0-12, 0 is 2, 12 is Ace
        :param card2_rank: 0-12, 0 is 2, 12 is Ace
        :param suited: True if the cards are the same suite, False otherwise
        :param player_count: number of players in the game
        :return: a dict with the following values: card_1_rank, card_2_rank, suited, win_rate, rank, percentile,
        player_count, sklansky, sklansky_position, modified_sklansky, modified_sklansky_position
        """
        if card1_rank == card2_rank and suited:
            raise ValueError("Cannot have the same card twice")
        if self.all_data is not None and not self.use_database:
            data = self.all_data.query(f"player_count == {player_count} and card_1_rank == {card1_rank} and "
                                       f"card_2_rank == {card2_rank} and {'' if suited else 'not '}suited")
        else:
            data = pd.read_sql(sqlalchemy.sql.text(f"select * from poker.win_rates where player_count = {player_count} and "
                                                   f"card_1_rank = '{card1_rank}' and card_2_rank = '{card2_rank}' "
                                                   f"and {'' if suited else 'not '}suited"), self.conn)
        return data.iloc[0].to_dict()
